/*
  mergemem.c
  Kernel module for 2.1.x (and in the future 2.2.x) Kernels
  
  This file is part of mergemem by Philipp Richter & Philipp Reisner

  mergemem is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
  
  mergemem is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with mergemem; see the file COPYING.  If not, write to
  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <asm/smp.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <linux/sched.h>

#include <asm/page.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/pgtable.h>

#include "mergemod.h"

#define MODULE_NAME "mergemod :"
#define MERGEMEM_INODE 3711

static ssize_t mergemod_read(struct file*, char*,size_t,loff_t *);
static ssize_t mergemod_write(struct file*, const char*,size_t,loff_t *);
static int mergemod(int, unsigned long, int, unsigned long );
static int genchksum(int, unsigned long, unsigned long* );

static struct file_operations proc_mergemod_file_operations = 
{
  NULL,                   /* lseek - default */
  mergemod_read,             /* read */
  mergemod_write,            /* write  */
  NULL,                   /* readdir */
  NULL,                   /* select - default */
  NULL,                   /* ioctl - default */
  NULL,                   /* mmap */
  NULL,                   /* no special open code */
  NULL,                   /* no special release code */
  NULL                    /* can't fsync */
};

static struct inode_operations proc_mergemod_inode_operations =
{
  &proc_mergemod_file_operations,   /* my own file-ops */
  NULL,                   /* create */
  NULL,                   /* lookup */
  NULL,                   /* link */
  NULL,                   /* unlink */
  NULL,                   /* symlink */
  NULL,                   /* mkdir */
  NULL,                   /* rmdir */
  NULL,                   /* mknod */
  NULL,                   /* rename */
  NULL,                   /* readlink */
  NULL,                   /* follow_link */
  NULL,                   /* readpage */
  NULL,                   /* writepage */
  NULL,                   /* bmap */
  NULL,                   /* truncate */
  NULL                    /* permission */
};
  

struct proc_dir_entry proc_mergemod = 
{
  MERGEMEM_INODE, 8, "mergemem",
  S_IFREG | S_IRUGO | S_IWUGO, 1, 0, 0,
  0, &proc_mergemod_inode_operations,
  NULL, NULL,
  NULL,
  NULL, NULL
};


int init_module()
{
  printk(MODULE_NAME"init_module()\n");

  proc_register(&proc_root,&proc_mergemod);
  
  return 0;
}

void cleanup_module()
{
  printk(MODULE_NAME"cleanup_module()\n");
  proc_unregister(&proc_root,MERGEMEM_INODE);
}

/****************************************************************************/
/* From here on, we have the interface to the user level process            */

char mergemod_test_message[] = "This is the kernel entry point of the mergemem"
                            "project\nthere is nothing of interest here\n";

static ssize_t mergemod_read(struct file* file, char* buf, size_t count,loff_t *unused)
{
  int fpos = file->f_pos;

  if(fpos > sizeof(mergemod_test_message)-1) return 0;
  if((sizeof(mergemod_test_message)-fpos) < count) 
    count=sizeof(mergemod_test_message)-fpos;
  copy_to_user(buf,mergemod_test_message+fpos,count);
  file->f_pos = file->f_pos+count;
  return count;
}

static ssize_t mergemod_write(struct file* file, const char* buf, size_t count,loff_t *unused)
{
  struct mergemod_struct mms;

  if(count != sizeof(struct mergemod_struct))
    {
      printk(MODULE_NAME"request of illegal size\n");
      return MERGEMEM_ILLREQ;
    }

  copy_from_user(&mms, buf, count);

  if(mms.pid2 == 0) 
    return genchksum(mms.pid1,mms.addr1,&mms.addr2);
  else  
    return mergemod(mms.pid1,mms.addr1,mms.pid2,mms.addr2);
}

/****************************************************************************/
/* From here on, we have the implementation of the page fusion function     */

static char * get_phys_addr(struct task_struct * tsk,unsigned long addr,
			    pte_t ** pe)
{
  pgd_t * page_dir;
  pmd_t * page_middle;
  pte_t pte;

  page_dir = pgd_offset(tsk->mm,addr);
  if (pgd_none(*page_dir)) return (char *)0;
  if (pgd_bad(*page_dir)) 
    {
      printk("Bad page dir entry %08lx\n", pgd_val(*page_dir));
      pgd_clear(page_dir);
      return (char *)0;
    }
  
  page_middle = pmd_offset(page_dir,addr);
  if (pmd_none(*page_middle)) return (char *)0;
  if (pmd_bad(*page_middle)) 
    {
      printk("Bad page middle entry %08lx\n", pmd_val(*page_middle));
      pmd_clear(page_middle);
      return (char *)0;
    }
  
  pte = *pte_offset(page_middle,addr);
  if (!pte_present(pte)) return (char *)0;

  *pe = pte_offset(page_middle,addr);

  return (char *) pte_page(pte);
}

static int mergemod(int pid1,unsigned long addr1,int pid2, unsigned long addr2)
{
  struct task_struct * tsk1, * tsk2;
  char * page1, * page2; 
  pte_t * pe1, * pe2;
  struct vm_area_struct * vma;
  
  /*  printk(MODULE_NAME"mergemod(%d,%lx;%d,%lx)\n",pid1,addr1,pid2,addr2); */

  tsk1 = find_task_by_pid(pid1);
  if(!tsk1) return MERGEMEM_NOTASK1;
  tsk2 = find_task_by_pid(pid2);
  if(!tsk2) return MERGEMEM_NOTASK2;
  
  page1 = get_phys_addr(tsk1,addr1,&pe1);
  if(!page1) return MERGEMEM_NOPAGE1;
  page2 = get_phys_addr(tsk2,addr2,&pe2);
  if(!page2) return MERGEMEM_NOPAGE2;
  
  if(page1 == page2) 
    return MERGEMEM_ALREADYSH;

  if(atomic_read(&mem_map[MAP_NR(page2)].count) != 1)
    return MERGEMEM_MOREMAP;

  if(memcmp(page1,page2,PAGE_SIZE))
     return MERGEMEM_NOTEQUAL;
  
  /* Do the actual work... */

  /*  set page1's pte to readonly */
  set_pte(pe1,pte_wrprotect(*pe1));

  /*  set page2's pte to page1 */
  /*  set page2's pte to readonly */
  vma = find_vma(tsk2->mm, addr2);
  set_pte(pe2,pte_wrprotect(mk_pte((unsigned long)page1,vma->vm_page_prot)));

  /*  increase the count in the according mem_map structure */
  atomic_inc(&mem_map[MAP_NR(page1)].count);

  /* page1 must go off the swap_cache, since it is now mapped more
   * than one time
   */
  delete_from_swap_cache(MAP_NR(page1));

  /*  free page2 */
  free_page((unsigned long)page2);

  /*  decrease the rss counters */
  tsk2->mm->rss--;
  
  /* No, TLB flushing needed, because the processes which PTEs where
   * altered, are not running at the moment.
   */
  
  return MERGEMEM_SUCCESS;
}

/* Calculate checksum for page */
static int genchksum(int pid, unsigned long addr, unsigned long* retval)
{
  struct task_struct * tsk;
  unsigned long * page, * end;
  pte_t * pe;
  unsigned long chksum;
    
  tsk = find_task_by_pid(pid);
  if(!tsk) return MERGEMEM_NOTASK1;
  
  page = (unsigned long *) get_phys_addr(tsk,addr,&pe);
  if(!page) return MERGEMEM_NOPAGE1;

  chksum=0;
  /* since this are unsigned long* pointers we must devide PAGE_SIZE
   * by the size of unsigned long (PAGE_SIZE is the size of a page in bytes)
   */
  /* For enhanced performance, we use only the first half of a page
   * for the checksum 
   */  
  end = page + (PAGE_SIZE/sizeof(unsigned long))/2;     
  while(page < end)
    {
      chksum = chksum ^ *page;
      page++;
    }
  *retval = chksum;
  return MERGEMEM_SUCCESS;
}
